/*
	COUNTDOWN LIVE - by Mattias Buelens
	File: ToolTip.js
	Desc: Class for tooltip handling
	Original by Eljay
*/

/* CONSTANTS */
//Maximum tooltip width in pixels
var TT_MAX_WIDTH = 400;

//ToolTip message constants
var WM_USER = 0x0400;
var TTM_SETTITLE = (WM_USER + 33);
var TTM_ADDTOOL = (WM_USER + 50);
var TTM_DELTOOL = (WM_USER + 51);
var TTM_GETTOOLINFO = (WM_USER + 53);
var TTM_SETTOOLINFO = (WM_USER + 54);
var TTM_UPDATETIPTEXT = (WM_USER + 57);
var TTM_SETMAXTIPWIDTH = (WM_USER + 24);
var TTF_SUBCLASS = 16;

//CreateWindow(Ex) constants
var WS_EX_TOPMOST = 0x8;
var WS_POPUP = 0x80000000;
var TTS_ALWAYSTIP = 0x1;
var TTS_NOPREFIX = 0x2;
var CW_USEDEFAULT = 0x80000000;

//SetWindowPos constants
var HWND_TOPMOST = -1;
var SWP_NOSIZE = 1;
var SWP_NOMOVE = 2;
var SWP_NOACTIVATE = 16;

//Tooltip icon constants
var TTI_NONE = 0;
var TTI_INFO = 1;
var TTI_WARNING = 2;
var TTI_ERROR = 3;
var TTI_INFO_LARGE = 4;
var TTI_WARNING_LARGE = 5;
var TTI_ERROR_LARGE = 6;

//Global object
var ToolTip;

/* ToolTip */
(function() {
	/* Private object stuff */
	var oToolTipIds = {};
	var oToolTipHandles = {};
	var oToolParents = {};
	var oDefaultCfg = {
		id: "*Unknown*",
		icon: TTI_NONE,
		title: "",
		tools: {}
	};

	//Clones an object to a new object
	function CloneObject(obj) {
		var newObj = {};
		for(var i in obj) newObj[i] = obj[i];
		return newObj;
	}
	//Extends an object
	function ExtendObject(obj, extend) {
		for(var i in extend) obj[i] = extend[i];
		return obj;
	}
	//Adds a tool to a ToolTip control
	function TT_AddTool(oToolTip, hTool, sText) {
		var TOOLINFO = Interop.Allocate(40);
		var Result = Interop.Call("user32", "GetClientRect", hTool, (TOOLINFO.DataPtr + 16));
		if(!Result) {
			TOOLINFO.Size = 0;
			throw new Error("ToolTip_GetClientRect");
			return false;
		}
		var lpText = Interop.Allocate(2*sText.length+2);
		lpText.WriteString(0, sText);
		with(TOOLINFO) {
			WriteDWORD(0, Size); // cbSize
			WriteDWORD(4, TTF_SUBCLASS); // uFlags
			WriteDWORD(8, hTool); // hWnd
			WriteDWORD(36, lpText.DataPtr);
		}
		Interop.Call("user32", "SendMessageW", oToolTip.Handle, TTM_ADDTOOL, 0, TOOLINFO);
		TOOLINFO.Size = 0;
		lpText.Size = 0;
		return true;
	}
	//Updates the tip text of a tool in a ToolTip control
	function TT_UpdateTipText(oToolTip, hTool, sText) {
		var TOOLINFO = Interop.Allocate(40);
		var lpText = Interop.Allocate(2*sText.length+2);
		lpText.WriteString(0, sText);
		with(TOOLINFO) {
			WriteDWORD(0, Size); // cbSize
			WriteDWORD(4, TTF_SUBCLASS); // uFlags
			WriteDWORD(8, hTool); // hWnd
			WriteDWORD(36, lpText.DataPtr);
		}
		Interop.Call("user32", "SendMessageW", oToolTip.Handle, TTM_UPDATETIPTEXT, 0, TOOLINFO);
		TOOLINFO.Size = 0;
		lpText.Size = 0;
		return true;
	}
	//Deletes a tool from a ToolTip control
	function TT_DeleteTool(oToolTip, hTool) {
		var TOOLINFO = Interop.Allocate(40);
		with(TOOLINFO) {
			WriteDWORD(0, Size); // cbSize
			WriteDWORD(4, TTF_SUBCLASS); // uFlags
			WriteDWORD(8, hTool); // hWnd
		}
		Interop.Call("user32", "SendMessageW", oToolTip.Handle, TTM_DELTOOL, 0, TOOLINFO);
		TOOLINFO.Size = 0;
		return true;
	}
	//Sets the title and icon of a ToolTip control
	function TT_SetTitle(oToolTip, sTitle, nIcon) {
		sTitle = sTitle ? (typeof sTitle === "string" ? sTitle : sTitle.toString()) : oDefaultCfg.title;
		nIcon = nIcon ? 1*nIcon : oDefaultCfg.icon;
		Interop.Call("user32", "SendMessageW", oToolTip.Handle, TTM_SETTITLE, nIcon, sTitle);
		return true;
	}

	/* ToolTip class */
	ToolTip = function(cfg) {
		//Messenger Plus! Live 4.21 or higher is required
		if(MsgPlus.Version < 4.21) {
			throw new Error("ToolTip_PlusVersion");
			return false;
		}

		//Get the configuration
		if(typeof cfg !== "object") return false;
		var priv = ExtendObject(CloneObject(oDefaultCfg), cfg);
		var oTools = {};

		//Create a tooltip window
		var hToolTip = Interop.Call("user32", "CreateWindowExW", WS_EX_TOPMOST, "tooltips_class32", 0, WS_POPUP | TTS_NOPREFIX | TTS_ALWAYSTIP, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, /*Parent*/0, 0, Interop.Call("kernel32", "GetCurrentProcess"), 0);
		if(!hToolTip) {
			Interop.Call("user32", "DestroyWindow", hToolTip);
			throw new Error("ToolTip_CreateWindowEx");
			return false;
		}
		//Position the tooltip window
		var Result = Interop.Call("user32", "SetWindowPos", hToolTip, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
		if(!Result) {
			Interop.Call("user32", "DestroyWindow", hTooltip);
			throw new Error("ToolTip_SetWindowPos");
			return false;
		}
		// Set the maximum width for the tooltip, this will wrap the text and enables the use of newline characters
		Interop.Call("user32", "SendMessageW", hToolTip, TTM_SETMAXTIPWIDTH, 0, TT_MAX_WIDTH);

		/* Public instance stuff */
		var Instance = {
			Handle: hToolTip,
			Id: priv.id,
			SetTitle: function(sTitle, nIcon) {
				priv.title = sTitle ? (typeof sTitle === "string" ? sTitle : sTitle.toString()) : priv.title;
				priv.icon = nIcon ? 1*nIcon : priv.icon;
				try {
					return TT_SetTitle(this, priv.title, priv.icon);
				} catch(e) {
					ToolTip.ErrorHandler(e);
					return false;
				}
			},
			Tool: function(hTool, sText) {
				hTool = 1*hTool;
				sText = sText ? (typeof sText === "string" ? sText : sText.toString()) : "";
				try {
					if(oTools[hTool]) {
						return TT_UpdateTipText(this, hTool, sText);
					} else {
						var Result = TT_AddTool(this, hTool, sText);
						if(Result) {
							oToolParents[hTool] = this;
							oTools[hTool] = true;
						}
						return Result;
					}
				} catch(e) {
					ToolTip.ErrorHandler(e);
					return false;
				}
			},
			DeleteTool: function(hTool) {
				hTool = 1*hTool;
				try {
					var Result = TT_DeleteTool(this, hTool);
				} catch(e) {
					ToolTip.ErrorHandler(e);
					return false;
				}
				if(Result) {
					delete oToolParents[hTool];
					delete oTools[hTool];
				}
				return Result;
			},
			Destroy: function() {
				var Result = Interop.Call("user32", "DestroyWindow", this.Handle);
				if(!Result) {
					throw new Error("ToolTip_Destroy");
					return false;
				}
				for(var hTool in oTools) {
					delete oToolParents[hTool];
				}
				delete oToolTipIds[this.Id];
				delete oToolTipHandles[this.Handle];
				return true;
			}
		};

		//Apply the configuration
		Instance.SetTitle();
		if(priv.tools) for(var hTool in priv.tools) {
			try {
				Instance.Tool(hTool, priv.tools[hTool]);
			} catch(e) { ToolTip.ErrorHandler(e); }
		}

		//Add this instance to the global indexes
		oToolTipIds[this.Id] = Instance;
		oToolTipHandles[this.Handle] = Instance;

		//Return the created instance
		return Instance;
	};
	
	/* Public object stuff */
	ToolTip.Icons = {
		"NONE": TTI_NONE,
		"INFO": TTI_INFO,
		"WARNING": TTI_WARNING,
		"ERROR": TTI_ERROR,
		"INFO_LARGE": TTI_INFO_LARGE,
		"WARNING_LARGE": TTI_WARNING_LARGE,
		"ERROR_LARGE": TTI_ERROR_LARGE
	};
	ToolTip.GetById = function(sId) {
		return (oToolTipIds[sId]) ? oToolTipIds[sId] : false;
	};
	ToolTip.GetByHandle = function(hToolTip) {
		hToolTip = 1*hToolTip;
		return (oToolTipHandles[hToolTip]) ? oToolTipHandles[hToolTip] : false;
	};
	ToolTip.GetByTool = function(hTool) {
		hTool = 1*hTool;
		return (oToolParents[hTool]) ? oToolParents[hTool] : false;
	};
	ToolTip.DestroyAll = function() {
		for(var hToolTip in oToolTipHandles) {
			try {
				oToolTipHandles[hToolTip].Destroy();
			} catch(e) { ToolTip.ErrorHandler(e); }
		}
	};
	ToolTip.ErrorHandler = function(e) {
		Debug.Trace("!!! ToolTip: Error occured (" + e.desc + ")");
	};

})();